import numpy as np

from seedbank import numpy_rng

def test_generator():
    rng = numpy_rng()
    assert isinstance(rng, np.random.Generator)


def test_generator_seed():
    rng = numpy_rng(42)
    assert isinstance(rng, np.random.Generator)


def test_generator_seed_seq():
    seq = np.random.SeedSequence(42)
    rng = numpy_rng(seq)
    assert isinstance(rng, np.random.Generator)

def test_generator_convert_from_rs():
    rng1 = np.random.RandomState()
    rng = numpy_rng(rng1)
    assert isinstance(rng, np.random.Generator)


def test_generator_passthrough():
    rng1 = np.random.default_rng()
    rng = numpy_rng(rng1)
    assert isinstance(rng, np.random.Generator)
    assert rng is rng1



import numpy as np
from seedbank import numpy_random_state

def test_random_state():
    rng = numpy_random_state()
    assert isinstance(rng, np.random.RandomState)


def test_random_state_seed():
    rng = numpy_random_state(42, )
    assert isinstance(rng, np.random.RandomState)


def test_random_state_passthrough():
    rng1 = np.random.RandomState()
    rng = numpy_random_state(rng1)
    assert isinstance(rng, np.random.RandomState)


def test_random_state_ss():
    seq = np.random.SeedSequence(42)
    rng = numpy_random_state(seq)
    assert isinstance(rng, np.random.RandomState)


def test_generator_convert_to_rs():
    rng1 = np.random.default_rng()
    rng = numpy_random_state(rng1)
    assert isinstance(rng, np.random.RandomState)



import os
import sys
sys.path.insert(0, os.path.abspath('..'))

import seedbank

project = 'SeedBank'
copyright = '2021 Boise State University'
author = 'Michael D. Ekstrand'

release = seedbank.__version__

extensions = [
    'sphinx.ext.napoleon',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
]

source_suffix = '.rst'

pygments_style = 'sphinx'
highlight_language = 'python3'

html_theme = 'furo'
html_theme_options = {
    # 'github_user': 'lenskit',
    # 'github_repo': 'seedbank',
}
templates_path = ['_templates']

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'numpy': ('https://docs.scipy.org/doc/numpy/', None),
    'numba': ('https://numba.readthedocs.io/en/stable/', None),
    'sklearn': ('https://scikit-learn.org/stable/', None),
    'torch': ('https://pytorch.org/docs/stable/', None),
}

autodoc_default_options = {
    'members': True,
    'member-order': 'bysource'
}

napoleon_type_aliases = {
    'seed-like': ':term:`seed-like`'
}



import logging
import numpy as np
try:
    from numba import njit
    _available = True
except ImportError:
    _available = False

_log = logging.getLogger(__name__)

if _available:
    @njit
    def _seed_numba(seed):
        np.random.seed(seed)


def is_available():
    return _available


def seed(state):
    _log.debug('initializing Numba seed')
    _seed_numba(state.int_seed)



from seedbank import derive_seed, initialize, _root_state, root_seed, int_seed
from seedbank._keys import make_key

def test_initialize():
    initialize(42)
    assert _root_state.seed.entropy == 42
    assert len(_root_state.seed.spawn_key) == 0
    assert root_seed().entropy == 42


def test_initialize_key():
    initialize(42, 'wombat')
    assert _root_state.seed.entropy == 42
    k = make_key('wombat')
    assert all(_root_state.seed.spawn_key[0] == k)
    assert root_seed().entropy == 42


def test_int_seed():
    initialize(42)
    assert int_seed() == _root_state.int_seed


def test_int_seed_words():
    initialize(42)
    words = int_seed(5)

    assert all(words == root_seed().generate_state(5))


def test_int_seed_new():
    initialize(42)
    seed = int_seed(seed=derive_seed())
    assert seed != int_seed()



import logging

try:
    import tensorflow as tf
except ImportError:
    tf = None

_log = logging.getLogger(__name__)


def is_available():
    if tf is None:
        return False
    elif hasattr(tf.random, 'set_seed'):
        return True
    else:
        _log.warning('TensorFlow 1 cannot be seeded')
        return False


def seed(state):
    seed = state.int_seed

    _tf_seed = getattr(tf.random, 'set_seed', None)
    if _tf_seed is not None:
        _log.debug('setting TensorFlow 2 seed')
        _tf_seed(seed)
    else:
        _log.warning('cannot set TensorFlow 1 seeds yet')



import numpy as np
import pytest
from seedbank import initialize

try:
    from numba import njit

    @njit
    def _numba_rand():
        return np.random.randint(10000)

except ImportError:
    pytestmark = pytest.mark.skip('Numba JIT not available')


def test_numba_init():
    initialize(42)
    a1 = _numba_rand()

    # if we re-initialize, should get the same result!
    initialize(42)
    a2 = _numba_rand()

    assert a1 == a2



"""
Common infrastructure for initializing random number generators.
"""

__version__ = '0.1.0'

import logging
from importlib import import_module
import numpy as np

from seedbank._state import SeedState
from seedbank._keys import make_seed

_log = logging.getLogger(__name__)
_root_state = SeedState()

__all__ = [
    'initialize',
    'derive_seed',
    'numpy_rng',
    'numpy_random_state'
]

# This list contains the modules that initialize seeds.
SEED_INITIALIZERS = [
    'seedbank.stdlib',
    'seedbank.numpy',
    'seedbank.numba',
    'seedbank.tensorflow',
    'seedbank.torch',
]


def initialize(seed, *keys):
    """
    Initialize the random infrastructure with a seed.  This function should generally be
    called very early in the setup.  This initializes all known and available RNGs with
    a seed derived from the specified seed.

    If you do **not** call this function, a default root seed is used, so functions like
    :func:`derive_seed` and :func:`numba_rng` work, but all other random number generators
    are left to their own default seeding behavior.

    Args:
        seed(int or str or numpy.random.SeedSequence):
            The random seed to initialize with.
        keys:
            Additional keys, to use as a ``spawn_key`` on the seed sequence.
            Passed to :func:`derive_seed`.
    Returns:
        numpy.random.SeedSequence:
            The random seed.
    """
    _root_state.initialize(seed, keys)
    _log.info('initialized root seed %s', _root_state.seed)

    for mod in SEED_INITIALIZERS:
        if isinstance(mod, str):
            mod = import_module(mod)
        if mod.is_available():
            mod.seed(_root_state)

    return _root_state.seed


def derive_seed(*keys, base=None):
    """
    Derive a seed from the root seed, optionally with additional seed keys.

    Args:
        keys(list of int or str):
            Additional components to add to the spawn key for reproducible derivation.
            If unspecified, the seed's internal counter is incremented (by calling
            :meth:`numpy.random.SeedSequence.spawn`).
        base(numpy.random.SeedSequence):
            The base seed to use.  If ``None``, uses the root seed.

    Returns:
        numpy.random.SeedSequence:
            The random seed.
    """
    return _root_state.derive(base, keys).seed


def root_seed():
    """
    Get the current root seed.

    Returns:
        numpy.random.SeedSequence:
            The root seed.
    """
    return _root_state.seed


def int_seed(words=None, seed=None):
    """
    Get the current root seed as an integer.

    Args:
        words(int or None):
            The number of words of entropy to return, or ``None`` for a single integer.
        seed(numpy.random.SeedSequence or None):
            The seed; if ``None``, returns the root seed.

    Returns:
        int or numpy.ndarray:
            The seed entropy.
    """

    if seed is None:
        seed = _root_state.seed

    if words is None:
        return seed.generate_state(1)[0]
    else:
        return seed.generate_state(words)


def numpy_rng(spec=None):
    """
    Get a NumPy random number generator.  This is similar to :func:`sklearn.utils.check_random_state`, but
    it returns a :class:`numpy.random.Generator` instead.

    Args:
        spec:
            The spec for this RNG.  Can be any of the following types:

            * ``int``
            * ``None``
            * :class:`numpy.random.SeedSequence`
            * :class:`numpy.random.RandomState` (its bit-generator is extracted and wrapped in a generator)
            * :class:`numpy.random.Generator` (returned as-is)

    Returns:
        numpy.random.Generator: A random number generator.
    """

    if isinstance(spec, np.random.Generator):
        return spec
    elif isinstance(spec, np.random.RandomState):
        return np.random.Generator(spec._bit_generator)
    elif spec is None:
        return np.random.default_rng(derive_seed())
    else:
        seed = make_seed(spec)
        return np.random.default_rng(seed)


def numpy_random_state(spec=None):
    """
    Get a legacy NumPy random number generator (:class:`numpy.random.mtrand.RandomState`).
    This is similar to :func:`sklearn.utils.check_random_state`.

    Args:
        spec:
            The spec for this RNG.  Can be any of the following types:

            * ``int``
            * ``None``
            * :class:`numpy.random.SeedSequence`
            * :class:`numpy.random.mtrand.RandomState`
            * :class:`numpy.random.Generator`

    Returns:
        numpy.random.mtrand.RandomState: A random number generator.
    """

    if isinstance(spec, np.random.RandomState):
        return spec
    elif isinstance(spec, np.random.Generator):
        return np.random.RandomState(spec._bit_generator)
    elif spec is None:
        return np.random.RandomState(int_seed(seed=derive_seed()))
    else:
        seed = make_seed(spec)
        return np.random.RandomState(int_seed(seed=seed))



import logging

try:
    import torch
except ImportError:
    torch = None

_log = logging.getLogger(__name__)


def is_available():
    return torch is not None


def seed(state):
    seed = state.int_seed
    torch.manual_seed(seed)



import hashlib
import numpy as np


def make_key(data):
    """
    Get a key, usable as entropy in a seed sequence, from a piece of data.
    """
    if isinstance(data, int) or isinstance(data, np.integer):
        return data
    if isinstance(data, np.ndarray):
        return data.astype(np.uint32)
    if isinstance(data, bytes):
        h = hashlib.md5(data)
        return np.frombuffer(h.digest(), np.uint32)
    if isinstance(data, str):
        return make_key(str.encode('utf8'))

    dt = type(data)
    raise TypeError(f'invalid seed type {dt}')


def make_seed(data):
    """
    Get a seed sequence from a piece of data.
    """
    if isinstance(data, np.random.SeedSequence):
        return data

    entropy = make_key(data)
    return np.random.SeedSequence(entropy)



import numpy as np

from seedbank._keys import make_key, make_seed


class SeedState:
    """
    Manage a root seed and facilities to derive seeds.
    """

    _seed: np.random.SeedSequence

    def __init__(self, seed=None):
        if seed is None:
            seed = np.random.SeedSequence()
        self._seed = seed

    @property
    def seed(self) -> np.random.SeedSequence:
        "Get the seed sequence for this seed state."
        return self._seed

    @property
    def int_seed(self):
        "Get this seed as an integer."
        return self.entropy(1)[0]

    def entropy(self, words):
        """
        Get *n* words of entropy as a NumPy array.

        Args:
            words(int): the number of words to return.

        Returns:
            numpy.ndarray: the entropy.
        """
        return self._seed.generate_state(words)

    def initialize(self, seed, keys):
        seed = make_seed(seed)

        if keys:
            seed = self.derive(seed, keys).seed

        self._seed = seed
        return seed

    def derive(self, base, keys=None):
        """
        Derive a new seed state.

        Args:
            base(seed-like):
                The base seed.  If ``None``, use this seed state.
            keys(list of seed-like):
                Additional keys for deriving the seed.  If no keys are
                provided, calls :meth:`numpy.random.SeedSequence.spawn` to
                obtain a new RNG.

        Returns:
            SeedState: the derived seed state.
        """
        if base is None:
            base = self.seed
        else:
            base = make_seed(base)

        if keys:
            k2 = tuple(make_key(k) for k in keys)
            seed = np.random.SeedSequence(base.entropy, spawn_key=base.spawn_key + k2)
        else:
            seed = base.spawn(1)[0]

        return SeedState(seed)

    def rng(self, seed=None):
        if seed is None:
            seed, = self.seed.spawn(1)
        elif not isinstance(seed, np.random.SeedSequence):
            seed = np.random.SeedSequence(make_key)
        return np.random.default_rng(seed)



from seedbank import initialize, derive_seed
from seedbank._keys import make_key

def test_derive_seed():
    initialize(42)
    s2 = derive_seed()
    assert s2.entropy == 42
    assert s2.spawn_key == (0,)


def test_derive_seed_intkey():
    initialize(42)
    s2 = derive_seed(10, 7)
    assert s2.entropy == 42
    assert s2.spawn_key == (10, 7)


def test_derive_seed_str():
    initialize(42)
    s2 = derive_seed(b'wombat')
    assert s2.entropy == 42
    assert all(s2.spawn_key[0] == make_key(b'wombat'))



import logging
import numpy as np

_log = logging.getLogger(__name__)


def is_available():
    return True


def seed(state):
    _log.debug('initializing NumPy root RNG')
    np.random.seed(state.int_seed)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['seedbank']

package_data = \
{'': ['*']}

install_requires = \
['numpy >=1.17']

extras_require = \
{'dev': ['flit', 'flake8', 'ipython', 'sphinx-autobuild'],
 'doc': ['sphinx', 'furo'],
 'numba': ['numba'],
 'test': ['pytest >= 6', 'pytest-doctestplus', 'pytest-cov'],
 'tf': ['tensorflow >=2,<3'],
 'torch': ['torch']}

setup(name='seedbank',
      version='0.1.0',
      description='Common infrastructure for initializing random number generators.',
      author='Michael Ekstrand',
      author_email='michaelekstrand@boisestate.edu',
      url='https://seedbank.lensit.org',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>= 3.7',
     )



"""
Environment management tool to instantiate Conda environments from Flit.
Requires flit-core and packaging to be installed.
"""

import os
import sys
import tempfile
import subprocess
from pathlib import Path
import argparse
from flit_core.config import read_flit_config, toml
from packaging.requirements import Requirement
from packaging.markers import default_environment


def write_env(obj, out):
    try:
        import yaml
        yaml.safe_dump(obj, out)
    except ImportError:
        import json
        json.dump(obj, out, indent=2)


def parse_args():
    parser = argparse.ArgumentParser(description='Manage development environments.')
    parser.add_argument('--python-version', '-V', metavar='VER',
                        help='use Python version VER')
    parser.add_argument('--extra', '-E', metavar='EXTRA', action='append',
                        help='include EXTRA')
    parser.add_argument('--name', '-n', metavar='NAME',
                        help='name Conda environment NAME')
    parser.add_argument('--no-dev', action='store_true', help='skip dev dependencies')
    parser.add_argument('--save-env', metavar='FILE',
                        help='save environment to FILE')
    parser.add_argument('--create-env', action='store_true',
                        help='create Conda environment')
    parser.add_argument('--update-env', action='store_true',
                        help='update Conda environment')
    args = parser.parse_args()
    return args


def load_project():
    tp = Path('pyproject.toml')
    fc = read_flit_config(tp)
    pyp = toml.loads(tp.read_text())
    return pyp, fc


class conda_config:
    def __init__(self, project):
        cfg = project.get('tool', {})
        cfg = cfg.get('envtool', {})
        self.config = cfg.get('conda', {})

    @property
    def name(self):
        return str(self.config.get('name', 'dev-env'))

    @property
    def channels(self):
        return [str(c) for c in self.config.get('channels', [])]

    @property
    def extras(self):
        return self.config.get('extras', {})

    def get_override(self, dep):
        ovr = self.config.get('overrides', {})
        dep_over = ovr.get(dep, {})
        if isinstance(dep_over, str):
            dep_over = {'name': dep_over}
        return dep_over

    def source(self, dep):
        dov = self.get_override(dep)
        return dov.get('source', None)

    def conda_name(self, dep):
        dov = self.get_override(dep)
        return str(dov.get('name', dep))


def marker_env(args):
    "Get the marker environment"
    env = {}
    env.update(default_environment())
    if args.python_version:
        env['python_version'] = args.python_version
        env['python_full_version'] = args.python_version
    return env


def req_active(env, req):
    if req.marker:
        return req.marker.evaluate(env)
    else:
        return True


def dep_str(cfg, req):
    dep = cfg.conda_name(req.name)
    if req.specifier:
        dep += f' {req.specifier}'
    return dep


def conda_env(args, pyp, flp):
    cfg = conda_config(pyp)
    mkenv = marker_env(args)
    name = args.name
    if name is None:
        name = cfg.name

    env = {'name': name}
    if cfg.channels:
        env['channels'] = cfg.channels

    deps = []
    if args.python_version:
        deps.append(f'python ={args.python_version}')
    elif flp.metadata['requires_python']:
        deps.append('python ' + str(flp.metadata['requires_python']))
    deps.append('pip')

    extras = set(['.none'])
    if not args.no_dev:
        extras |= set(['dev', 'doc', 'test'])
    if args.extra:
        for e in args.extra:
            if e == 'all':
                extras |= set(flp.reqs_by_extra.keys())
            else:
                extras.add(e)

    pip_deps = []

    for e in extras:
        for req in flp.reqs_by_extra.get(e, []):
            req = Requirement(req)
            if req_active(mkenv, req):
                if req.url or cfg.source(req.name) == 'pip':
                    pip_deps.append(req)
                else:
                    deps.append(dep_str(cfg, req))
        for cr in cfg.extras.get(e, []):
            deps.append(str(cr))

    if pip_deps:
        deps.append({'pip': [str(r) for r in pip_deps]})
    env['dependencies'] = deps

    return env


def env_command(env, cmd):
    with tempfile.TemporaryDirectory() as td:
        path = Path(td)
        ef = path / 'environment.yml'
        with ef.open('w') as f:
            write_env(env, f)
        print(cmd, 'environment', ef)
        subprocess.run(['conda', 'env', cmd, '-f', os.fspath(ef)], check=True)


def main(args):
    py_p, flit_p = load_project()
    env = conda_env(args, py_p, flit_p)
    if args.save_env:
        with open(args.save_env, 'w') as ef:
            write_env(env, ef)
    elif args.create_env:
        env_command(env, 'create')
    elif args.update_env:
        env_command(env, 'update')
    else:
        write_env(env, sys.stdout)


if __name__ == '__main__':
    main(parse_args())
